// Louis-Clément LEFÈVRE
//
// Modèle pour les TP d'informatique

#import "hoche.typ": *

#let hoche-TP(
  numéro: none,
  titre: none,
  keywords: none,
  body
) = {
  let num = if (numéro != none) { if numéro < 10 { "0" + str(numéro) } else { str(numéro) } }
  let titre-page = "TP" + if (numéro != none) { sym.space.nobreak + str(numéro)} + if (titre != none) { linebreak() + titre }
  let titre-header = titre
  let titre-PDF = if (numéro != none) { "TP" + " " + num + " : " + titre}
  let code = if (numéro != none) { "TP" + " " + num }
  show: hoche.with(
    titre: titre-page,
    titre-header: titre,
    titre-PDF: titre-PDF,
    code: code,
    numéro-page: true,
    no-header-first-page: true,
    extra-keywords: ("informatique", ) + keywords,
  )

  // style latex
  set par(leading: 0.55em)
  show heading: set block(above: 1.4em, below: 1em)
  show heading: it => block( context {
    counter(heading).display() + h(1em) + it.body
  })

  show heading.where(level: 3): it => {
    strong(it.body) + h(2em)
  }

  // code
  let _block_raw(it) = {
    block(
      breakable: false,
      fill: silver.lighten(50%),
      radius: 6pt,
      inset: 6pt,
      width: 100%,
      it
    )
  }
  // le thème qui va bien
  set raw(theme: "Pastie.tmTheme")
  show raw.where(block: true): _block_raw
  show raw.where(block: true, lang: "pycon"): it => _block_raw({
    it.text.split("\n").map(ligne => {
      if ligne.starts-with(">>>") { raw(lang: "python", ligne) }
      else if ligne.starts-with(regex("(AssertionError)|(KeyError)|(IndexError)|(TypeError)|(ZeroDivisionError)\\:")) {
        set text(fill: red)
        raw(lang: "text", ligne)
      } else {
        let i = ligne.position("#")
        if i == none {
          raw(lang: "text", ligne)
        } else {
          let a = ligne.slice(0, i)
          let b = ligne.slice(i)
          raw(lang: "text", a) + raw(lang: "python", b)
        }
      }
    }).join(linebreak())
  })
  show raw.where(block: false, lang: "pycon"): it => raw(lang: "python", it.text)

  show table: set block(breakable: false)
  set table(
    stroke: 0.4pt,
    inset: 3pt,
  )

  set enum(
    number-align: start+top,
    body-indent: 0em,
    full: true,
    numbering: (..n) => {
      if n.pos().len() == 1 {
        box(width: 18pt, numbering("1.", ..n))
      } else if n.pos().len() == 2 {
        align(center,
          numbering("(i)", ..(n.pos().slice(1))) + h(3pt)
        )
      } else {
        numbering("1.", ..(n.pos().slice(2))) + h(3pt)
      }
    },
  )

  body
}

// style d'exercices
#let counter_exercice = counter("exercice")
#let exercice(
  soustitre: none,
  difficulté: none,
  body
) = {
  counter_exercice.step()
  block(
    inset: 6pt,
    width: 100%,
    radius: 6pt,
    stroke: 0.4pt + black, {
      context {
        block(below: 0.55em, sticky: true, {
          strong("Exercice" + sym.space.nobreak + counter_exercice.display())
          if difficulté != none {
            sym.space
            strong(
              "(" + for _ in range(difficulté) { "*" } + ")"
            )
          }
          if soustitre != none { h(1em) + emph(soustitre)}
        })
      }
      body
    }
  )
}

#let thmqed = {
  h(1fr) + sym.wj + sym.space.nobreak + $square.stroked$
}

#let thmtheorem(body) = {
  block(
    width: 100%,
    inset: 6pt,
    radius: 6pt,
    stroke: (thickness: 0.4pt, dash: "dashed"),
    {
      [*Théorème.*] + [ ]
      emph(body)
    }
  )
}

#let thmdefinition(body) = {
  block(
    width: 100%,
    inset: 6pt,
    radius: 6pt,
    stroke: (thickness: 0.4pt, dash: "dashed"),
    {
      [*Définition.*] + [ ]
      body
    }
  )
}

#let thmproof(qed: true, body) = {
  block(
    width: 100%,
    {
      [_Démonstration._] + [ ]
      body
      if qed { thmqed }
    }
  )
}

#let thmremark(body) = {
  block(
    width: 100%,
    inset: 6pt,
    radius: 6pt,
    stroke: (thickness: 0.4pt, dash: "dashed"),
    {
      [_Remarque._] + [ ]
      body
    }
  )
}

#let _boitecouleur(couleur, message, body) = {
  block(
    width: 100%,
    radius: 6pt,
    breakable: false,
    stroke: 1pt+couleur,
    fill: couleur.lighten(90%),
    {
      block(
        sticky: true,
        width: 100%,
        inset: 6pt,
        fill: couleur,
        radius: (top: 6pt),
        below: 0pt,
        text(fill: white, strong(message))
      )
      pad(6pt, body)
    }
  )
}

#let boiteavertissement(body) = {
  _boitecouleur(red, "Avertissement", body)
}

#let boitearetenir(body) = {
  _boitecouleur(blue, "À retenir", body)
}
